package	jp.co.sds_corp.util.jmx.rmi.server;

import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Hashtable;
import java.util.Set;
import java.util.Vector;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.InstanceAlreadyExistsException;
import javax.management.InstanceNotFoundException;
import javax.management.IntrospectionException;
import javax.management.InvalidAttributeValueException;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.NotCompliantMBeanException;
import javax.management.Notification;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectInstance;
import javax.management.ObjectName;
import javax.management.QueryExp;
import javax.management.ReflectionException;

import jp.co.sds_corp.util.jmx.rmi.RMIConnector;
import jp.co.sds_corp.util.jmx.rmi.RMINotificationListener;

public class RMIConnectorImpl extends UnicastRemoteObject
	implements RMIConnector
{
	private MBeanServer					mServer;
	private Hashtable<Object,Object>	mObjectPool = new Hashtable<Object,Object>();
	private Vector<NotificationListener>	mListeners = new Vector<NotificationListener>();

	public RMIConnectorImpl( MBeanServer pServer ) throws RemoteException {
		super();
		mServer = pServer;
	}

	public Object instantiate( String pClassName )
		throws ReflectionException, MBeanException, RemoteException
	{
		return assignObjectHandler( mServer.instantiate( pClassName ) );
	}

	public Object instantiate( String pClassName, ObjectName pLoaderName )
		throws ReflectionException, MBeanException,
				InstanceNotFoundException, RemoteException
	{
		return assignObjectHandler( mServer.instantiate( pClassName, pLoaderName ) );
	}

	public Object instantiate( String pClassName, Object[] pParams,
								String[] pSignature )
		throws ReflectionException, MBeanException, RemoteException
	{
		checkForObjectHandlers( pParams, pSignature );
		return assignObjectHandler(
			mServer.instantiate( pClassName, pParams, pSignature )
		);
	}

	public Object instantiate( String pClassName, ObjectName pLoaderName,
					Object[] pParams, String[] pSignature )
		throws ReflectionException, MBeanException,
				InstanceNotFoundException, RemoteException
	{
		checkForObjectHandlers( pParams, pSignature );
		return assignObjectHandler(
			mServer.instantiate( pClassName, pLoaderName, pParams, pSignature )
		);
	}

	public ObjectInstance createMBean( String pClassName, ObjectName pName )
		throws ReflectionException, InstanceAlreadyExistsException,
				MBeanRegistrationException, MBeanException,
				NotCompliantMBeanException, RemoteException
	{
		return mServer.createMBean( pClassName, pName );
	}

	public ObjectInstance createMBean( String pClassName, ObjectName pName,
					ObjectName pLoaderName )
		throws ReflectionException, InstanceAlreadyExistsException,
				MBeanRegistrationException, MBeanException,
				NotCompliantMBeanException, InstanceNotFoundException,
				RemoteException
	{
		return mServer.createMBean( pClassName, pName, pLoaderName );
	}

	public ObjectInstance createMBean( String pClassName, ObjectName pName,
					Object[] pParams, String[] pSignature )
		throws
				ReflectionException, InstanceAlreadyExistsException,
				MBeanRegistrationException, MBeanException,
				NotCompliantMBeanException, RemoteException
	{
		checkForObjectHandlers( pParams, pSignature );
		return mServer.createMBean( pClassName, pName, pParams, pSignature );
	}

	public ObjectInstance createMBean( String pClassName, ObjectName pName,
		ObjectName pLoaderName, Object[] pParams, String[] pSignature )
		throws ReflectionException, InstanceAlreadyExistsException,
				MBeanRegistrationException, MBeanException,
				NotCompliantMBeanException, InstanceNotFoundException,
				RemoteException
	{
		checkForObjectHandlers( pParams, pSignature );
		return mServer.createMBean( pClassName, pName, pLoaderName, pParams, pSignature );
	}

	public ObjectInstance registerMBean( Object pObjectHandler, ObjectName pNameToAssign )
		throws InstanceAlreadyExistsException, MBeanRegistrationException,
				NotCompliantMBeanException, RemoteException
	{
		if( !( pObjectHandler instanceof ObjectHandler ) ) {
			throw new IllegalArgumentException(
				"You can only register local objects referenced by ObjectHandler"
			);
		}
		return mServer.registerMBean(
			checkForObjectHandler( pObjectHandler ), pNameToAssign
		);
	}

	public void unregisterMBean( ObjectName pName )
		throws InstanceNotFoundException, MBeanRegistrationException,
				RemoteException
	{
		mServer.unregisterMBean( pName );
	}

	public ObjectInstance getObjectInstance( ObjectName pName )
		throws InstanceNotFoundException, RemoteException
	{
		return mServer.getObjectInstance( pName );
	}

	public Set<ObjectInstance> queryMBeans( ObjectName pName, QueryExp pQuery )
		throws RemoteException
	{
		return mServer.queryMBeans( pName, pQuery );
	}

	public Set<ObjectName> queryNames( ObjectName pName, QueryExp pQuery )
		throws RemoteException
	{
		return mServer.queryNames( pName, pQuery );
	}

	public boolean isRegistered( ObjectName pName )
		throws RemoteException
	{
		return mServer.isRegistered( pName );
	}

	public boolean isInstanceOf( ObjectName pName, String pClassName )
		throws InstanceNotFoundException, RemoteException
	{
		return mServer.isInstanceOf( pName, pClassName );
	}

	public Integer getMBeanCount() throws RemoteException {
		return mServer.getMBeanCount();
	}

	public Object getAttribute( ObjectName pName, String pAttribute )
		throws MBeanException, AttributeNotFoundException,
				InstanceNotFoundException, ReflectionException,
				RemoteException
	{
		return mServer.getAttribute( pName, pAttribute );
	}

	public AttributeList getAttributes( ObjectName pName, String[] pAttributes )
		throws InstanceNotFoundException, ReflectionException, RemoteException
	{
		return mServer.getAttributes( pName, pAttributes );
	}

	public void setAttribute( ObjectName pName, Attribute pAttribute )
		throws InstanceNotFoundException, AttributeNotFoundException,
				InvalidAttributeValueException, MBeanException,
				ReflectionException, RemoteException
	{
		mServer.setAttribute( pName, pAttribute );
	}

	public AttributeList setAttributes( ObjectName pName, AttributeList pAttributes )
		throws InstanceNotFoundException, ReflectionException, RemoteException
	{
		return mServer.setAttributes( pName, pAttributes );
	}

	public Object invoke( ObjectName pName, String pActionName,
		Object[] pParams, String[] pSignature )
		throws InstanceNotFoundException, MBeanException,
				ReflectionException, RemoteException
	{
		checkForObjectHandlers( pParams, pSignature );
		return mServer.invoke( pName, pActionName, pParams, pSignature );
	}

	public String getDefaultDomain() throws RemoteException {
		return mServer.getDefaultDomain();
	}

	public void addNotificationListener( ObjectName pName,
		RMINotificationListener pListener,
		NotificationFilter pFilter, Object pHandback )
		throws InstanceNotFoundException, RemoteException
	{
		NotificationListener lRemoteListener = new Listener( pListener );
		mServer.addNotificationListener(
			pName, lRemoteListener, pFilter, pHandback
		);
		mListeners.addElement( lRemoteListener );
	}

	public void removeNotificationListener( ObjectName pName, RMINotificationListener pListener )
		throws InstanceNotFoundException, ListenerNotFoundException,
				RemoteException
	{
		int lIndex = mListeners.indexOf( new Listener( pListener ) );
		if( lIndex >= 0 ) {
			mServer.removeNotificationListener(
				pName, (Listener) mListeners.elementAt( lIndex )
			);
		}
	}

	public MBeanInfo getMBeanInfo( ObjectName pName )
		throws InstanceNotFoundException, IntrospectionException,
				ReflectionException, RemoteException
	{
		return mServer.getMBeanInfo( pName );
	}

	private void checkForObjectHandlers( Object[] pListOfObjects, String[] pSignature )
	{
		for( int i = 0; i < pListOfObjects.length; i++ ) {
			Object lEffective = checkForObjectHandler( pListOfObjects[ i ] );
			if( pListOfObjects[ i ] != lEffective ) {
				pListOfObjects[ i ] = lEffective;
				if( i < pSignature.length ) {
					if( pSignature[ i ].equals( ObjectHandler.class.getName() ) ) {
						pSignature[ i ] = lEffective.getClass().getName();
					}
				}
			}
		}
	}
	private Object checkForObjectHandler( Object pObjectToCheck ) {
		if( pObjectToCheck instanceof ObjectHandler ) {
			return mObjectPool.get( pObjectToCheck );
		} else {
			return pObjectToCheck;
		}
	}

	private ObjectHandler assignObjectHandler( Object pNewObject ) {
		ObjectHandler lObjectHandler = new ObjectHandler(
			this.toString()
		);
		mObjectPool.put(
			lObjectHandler,
			pNewObject
		);
		return lObjectHandler;
	}

	private class Listener implements NotificationListener {
		private RMINotificationListener		mRemoteListener;
		public Listener( RMINotificationListener pRemoteListener ) {
			mRemoteListener = pRemoteListener;
		}

		public void handleNotification( Notification pNotification, Object pHandback )
		{
			try {
				mRemoteListener.handleNotification( pNotification, pHandback );
			} catch( RemoteException re ) {
				re.printStackTrace();
			}
		}
		public boolean equals( Object pTest ) {
			if( pTest instanceof Listener ) {
				return mRemoteListener.equals( ( (Listener) pTest).mRemoteListener );
			}
			return false;
		}
		public int hashCode() {
			return mRemoteListener.hashCode();
		}
	}
}

